package job

import (
	"context"
	"fmt"
	"time"

	"github.com/mitchellh/mapstructure"
	"go.uber.org/zap"

	"github.com/freeman42332/zdb1000n/src/job/config"
	"github.com/freeman42332/zdb1000n/src/utils"
	"github.com/freeman42332/zdb1000n/src/utils/metrics"
	"github.com/freeman42332/zdb1000n/src/utils/templates"
)

func logJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (
	data any, err error, //nolint:unparam // data is here to match Job
) {
	var jobConfig struct {
		Text string
	}

	if err := mapstructure.Decode(args, &jobConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	logger.Info(templates.ParseAndExecute(logger, jobConfig.Text, ctx))

	return nil, nil
}

func setVarJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (data any, err error) {
	var jobConfig struct {
		Value string
	}

	if err := mapstructure.Decode(args, &jobConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	return templates.ParseAndExecute(logger, jobConfig.Value, ctx), nil
}

func checkJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (
	data any, err error, //nolint:unparam // data is here to match Job
) {
	var jobConfig struct {
		Value string
	}

	if err := mapstructure.Decode(args, &jobConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	if templates.ParseAndExecute(logger, jobConfig.Value, ctx) != "true" {
		return nil, fmt.Errorf("validation failed %v", jobConfig.Value)
	}

	return nil, nil
}

func sleepJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (data any, err error) {
	var jobConfig struct {
		Value time.Duration
	}

	if err := utils.Decode(args, &jobConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	time.Sleep(jobConfig.Value)

	return nil, nil
}

func discardErrorJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (
	data any, err error, //nolint:unparam // data is here to match Job
) {
	var jobConfig struct {
		BasicJobConfig

		Job config.Config
	}

	if err := ParseConfig(&jobConfig, args, *globalConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	job := Get(jobConfig.Job.Type)
	if job == nil {
		logger.Debug("unknown job, discarding", zap.String("job", jobConfig.Job.Type))

		return nil, nil
	}

	data, err = job(ctx, jobConfig.Job.Args, globalConfig, a, logger)
	if err != nil {
		logger.Debug("discarded error", zap.Error(err))
	}

	return data, nil
}

func timeoutJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (
	data any, err error, //nolint:unparam // data is here to match Job
) {
	var jobConfig struct {
		BasicJobConfig

		Timeout time.Duration
		Job     config.Config
	}

	if err := ParseConfig(&jobConfig, args, *globalConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	ctx, cancel := context.WithTimeout(ctx, jobConfig.Timeout)
	defer cancel()

	job := Get(jobConfig.Job.Type)

	if job == nil {
		return nil, fmt.Errorf("unknown job %q", jobConfig.Job.Type)
	}

	return job(ctx, jobConfig.Job.Args, globalConfig, a, logger)
}

func loopJob(ctx context.Context, args config.Args, globalConfig *GlobalConfig, a *metrics.Accumulator, logger *zap.Logger) (data any, err error) {
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	var jobConfig struct {
		BasicJobConfig

		Job config.Config
	}

	if err := ParseConfig(&jobConfig, args, *globalConfig); err != nil {
		return nil, fmt.Errorf("error parsing job config: %w", err)
	}

	for jobConfig.Next(ctx) {
		job := Get(jobConfig.Job.Type)
		if job == nil {
			return nil, fmt.Errorf("unknown job %q", jobConfig.Job.Type)
		}

		data, err := job(ctx, jobConfig.Job.Args, globalConfig, a, logger)
		if err != nil {
			return nil, fmt.Errorf("error running job: %w", err)
		}

		ctx = context.WithValue(ctx, templates.ContextKey("data."+jobConfig.Job.Name), data)
	}

	return nil, nil
}
